import numpy as np
import tensorflow as tf
from tensorflow.python.keras import layers

from easy_tpp.utils import py_assert
from easy_tpp.utils.tf_utils import get_shape_list

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


def gelu(x):
    """Gaussian Error Linear Unit.
    This is a smoother version of the RELU.
    Original paper: https://arxiv.org/abs/1606.08415
    Args:
        x: float Tensor to perform activation.
    Returns:
        `x` with the GELU activation applied.
    """
    cdf = 0.5 * (1.0 + tf.tanh((np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3)))))
    return x * cdf


def swish(x):
    return x * tf.sigmoid(x)


def null_activation(x):
    return x


def activation_layer(activation):
    if activation.lower() == 'null':
        return null_activation
    if activation.lower() == 'gelu':
        act_layer = gelu
    elif activation.lower() == 'swish':
        act_layer = swish
    elif isinstance(activation, str):
        act_layer = tf.keras.layers.Activation(activation)
    elif issubclass(activation, layers.Layer):
        act_layer = activation()
    else:
        raise ValueError(
            "Invalid activation,found %s.You should use a str or a Activation Layer Class." % (activation))
    return act_layer


def append_tensor_alias(tensor, alias):
    """Append an alias to the list of aliases of the tensor.
    Args:
      tensor: A `Tensor`.
      alias: String, to add to the list of aliases of the tensor.
    Returns:
      The tensor with a new alias appended to its list of aliases.
    """
    # Remove ending '/' if present.
    if alias[-1] == '/':
        alias = alias[:-1]
    if hasattr(tensor, 'aliases'):
        tensor.aliases.append(alias)
    else:
        tensor.aliases = [alias]
    return tensor


class LayerNormalization(layers.Layer):
    """

    ref: https://github.com/shenweichen/DeepCTR/blob/master/deepctr/layers/normalization.py

    """

    def __init__(self, axis=-1, eps=1e-9, center=True,
                 scale=True, **kwargs):
        self.axis = axis
        self.eps = eps
        self.center = center
        self.scale = scale
        super(LayerNormalization, self).__init__(**kwargs)

    def build(self, input_shape):
        self.gamma = self.add_weight(name='gamma', shape=input_shape[-1:],
                                     initializer=tf.keras.initializers.Ones(), trainable=True)
        self.beta = self.add_weight(name='beta', shape=input_shape[-1:],
                                    initializer=tf.keras.initializers.Zeros(), trainable=True)
        super(LayerNormalization, self).build(input_shape)

    def call(self, inputs):
        mean = tf.keras.backend.mean(inputs, axis=self.axis, keepdims=True)
        variance = tf.keras.backend.mean(tf.keras.backend.square(inputs - mean), axis=-1, keepdims=True)
        std = tf.keras.backend.sqrt(variance + self.eps)
        outputs = (inputs - mean) / std
        if self.scale:
            outputs *= self.gamma
        if self.center:
            outputs += self.beta
        return outputs


class MultiHeadAttention(layers.Layer):
    def __init__(self, hidden_size, num_heads, dropout_rate, **kwargs):
        super(MultiHeadAttention, self).__init__(**kwargs)
        self.hidden_size = hidden_size
        self.num_heads = num_heads
        py_assert(hidden_size % num_heads == 0, ValueError, 'In Attention, hidden size '
                                                            'must be a multiplier of num_heads')
        self.head_size = int(hidden_size / num_heads)
        self.dropout = layers.Dropout(dropout_rate)

    def build(self, input_shape):
        """Build up weight tensors.

        Args:
            input_shape (tensor): shape of the input.
        """
        self.q_head_weight = self.add_weight(
            shape=(self.hidden_size, self.head_size * self.num_heads),
            initializer=tf.keras.initializers.TruncatedNormal(),
            dtype=tf.float32,
            name='query/kernel')

        self.k_head_weight = self.add_weight(
            shape=(self.hidden_size, self.head_size * self.num_heads),
            initializer=tf.keras.initializers.TruncatedNormal(),
            dtype=tf.float32,
            name='key/kernel')

        self.v_head_weight = self.add_weight(
            shape=(self.hidden_size, self.head_size * self.num_heads),
            initializer=tf.keras.initializers.TruncatedNormal(),
            dtype=tf.float32,
            name='value/kernel')

        self.q_head_bias = self.add_weight(
            shape=(self.head_size * self.num_heads,),
            initializer=tf.keras.initializers.TruncatedNormal(),
            dtype=tf.float32,
            name='query/bias')

        self.k_head_bias = self.add_weight(
            shape=(self.head_size * self.num_heads,),
            initializer=tf.keras.initializers.TruncatedNormal(),
            dtype=tf.float32,
            name='key/bias')

        self.v_head_bias = self.add_weight(
            shape=(self.head_size * self.num_heads,),
            initializer=tf.keras.initializers.TruncatedNormal(),
            dtype=tf.float32,
            name='value/bias')

        super(MultiHeadAttention, self).build(input_shape)

    def _abs_attn_core(self, q_head, k_head, v_head, attn_mask, training,
                       scale):
        """Compute the attention weight score.

        Args:
            q_head (tensor): [batch_size, q_seq_len, num_head, head_size].
            k_head (tensor): [batch_size, kv_seq_len, num_head, head_size].
            v_head (tensor): [batch_size, kv_seq_len, num_head, head_size].
            attn_mask (tensor): [batch_size, q_seq_len, kv_seq_len].
            training (bool): whether in training mode.
            scale (float): equals to 1 / (head_size ** 0.5), which scales down the attention score.

        Returns:
            tuple: attention context vector, [batch_size, q_seq_len, num_head, head_size];
             and attention weight [batch_size, num_head, q_seq_len, kv_seq_len].
        """
        # [batch_size, num_head, q_seq_len, kv_seq_len]
        attn_score = tf.einsum('bind,bjnd->bnij', q_head, k_head)

        # [batch_size, num_head, q_seq_len, kv_seq_len]
        attn_score = tf.multiply(attn_score, scale)

        # [batch_size, 1, q_seq_len, kv_seq_len]
        attn_mask = tf.expand_dims(attn_mask, axis=[1])

        # elements 1 => need to mask, 0 => no mask
        # adder = (1.0 - tf.cast(attn_mask, tf.float32)) * -10000.0
        adder = tf.cast(attn_mask, tf.float32) * -10000.0
        attn_score += adder

        # [batch_size, num_head, q_seq_len, kv_seq_len]
        attn_prob = tf.nn.softmax(attn_score)
        attn_prob = self.dropout(attn_prob, training=training)

        # [batch_size, q_seq_len, num_head, head_size]
        attn_vec = tf.einsum('bnij,bjnd->bind', attn_prob, v_head)
        return attn_vec, attn_prob

    def call(self, attention_input, attention_mask, kv=None, training=False, output_weight=False, **kwargs):

        q_input = attention_input
        if kv is None:
            k_input = attention_input
            v_input = attention_input
        else:
            k_input = v_input = kv

        batch_size, q_seq_length, kv_seq_length = get_shape_list(attention_mask)

        q_head_h = tf.einsum('bih,hx->bix', q_input, self.q_head_weight)
        q_head_h = tf.nn.bias_add(q_head_h, self.q_head_bias)

        k_head_h = tf.einsum('bih,hx->bix', k_input, self.k_head_weight)
        k_head_h = tf.nn.bias_add(k_head_h, self.k_head_bias)

        v_head_h = tf.einsum('bih,hx->bix', v_input, self.v_head_weight)
        v_head_h = tf.nn.bias_add(v_head_h, self.v_head_bias)

        q_head_h = tf.reshape(q_head_h, [batch_size, q_seq_length, self.num_heads, self.head_size])
        k_head_h = tf.reshape(k_head_h, [batch_size, kv_seq_length, self.num_heads, self.head_size])
        v_head_h = tf.reshape(v_head_h, [batch_size, kv_seq_length, self.num_heads, self.head_size])

        scale = 1 / (self.head_size ** 0.5)
        attn_vec, attn_prob = self._abs_attn_core(q_head_h, k_head_h, v_head_h, attention_mask, training, scale)
        attn_vec = tf.reshape(attn_vec, [batch_size, q_seq_length, self.head_size * self.num_heads])
        if output_weight:
            return attn_vec, attn_prob
        else:
            return attn_vec


class DenseDropoutLayernorm(layers.Layer):
    def __init__(self, hidden_size, dropout_rate, **kwargs):
        super(DenseDropoutLayernorm, self).__init__(**kwargs)
        self.dense = layers.Dense(hidden_size)
        self.LayerNorm = LayerNormalization()
        self.dropout = layers.Dropout(dropout_rate)

    def call(self, inputs, training=False):
        hidden_states, input_tensor = inputs
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states, training=training)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class EncoderLayer(layers.Layer):
    def __init__(self, hidden_size, num_heads, dropout_rate, **kwargs):
        super(EncoderLayer, self).__init__(**kwargs)
        self.input_layer = layers.Dense(hidden_size)
        self.self_attention = MultiHeadAttention(hidden_size, num_heads, dropout_rate)
        self.dense_output = DenseDropoutLayernorm(hidden_size, dropout_rate)

    def call(self, inputs, training=False):
        input_tensor, attention_mask = inputs
        input_tensor = self.input_layer(input_tensor)
        self_outputs = self.self_attention(input_tensor, attention_mask, training=training)
        attention_output = self.dense_output([self_outputs, input_tensor], training=training)
        return attention_output


class TimePositionalEncoding(layers.Layer):
    def __init__(self, hidden_size, max_len=5000):
        """Standard positional encoder,
        source code is from https://www.tensorflow.org/text/tutorials/transformer#the_encoder_layer
        Args:
            hidden_size (int):
            max_len:
        """
        super(TimePositionalEncoding, self).__init__()

        depth = hidden_size / 2
        # [max_len, 1]
        positions = np.arange(max_len)[:, None]
        # [1, depth]
        depths = np.arange(depth)[None, :] / depth
        # [1, depth]
        angle_rates = 1 / (10000 ** depths)
        # [max_len, depth]
        self.angle_rads = positions * angle_rates

        pos_encoding = np.concatenate(
            [np.sin(self.angle_rads), np.cos(self.angle_rads)],
            axis=-1)

        # [1, max_len, model_dim]
        self.pe = tf.cast(pos_encoding, dtype=tf.float32)[None, ...]

    def call(self, inputs, **kwargs):
        """Compute time positional encoding defined in Equation (2) in THP model.

        Args:
            inputs: (tensor), [batch_size, seq_len]
            **kwargs:

        Returns:
            positional encoding

        """

        _, seq_len = get_shape_list(inputs)

        return self.pe[:, :seq_len]


class TimeShiftedPositionalEncoding(TimePositionalEncoding):
    def __init__(self, hidden_size, max_len=5000):
        super(TimeShiftedPositionalEncoding, self).__init__(hidden_size, max_len)
        self.hidden_size = hidden_size

        self.layer_time_delta = layers.Dense(hidden_size // 2)

        depth = self.hidden_size / 2

        # [1, depth]
        self.depths = tf.cast(tf.range(depth)[None, :] / depth, tf.float32)

    def call(self, inputs, **kwargs):
        """Compute time shifted positional encoding defined in Equation (8) in SAHP model

        Args:
            inputs: (time_seqs, time_delta_seqs), [batch_size, seq_len]
            **kwargs:

        Returns:
            positional encoding

        """
        time_seqs, time_delta_seqs = inputs
        _, seq_len = get_shape_list(time_seqs)

        # [batch_size, seq_len, model_dim //2]
        phi = self.layer_time_delta(time_delta_seqs[..., None])

        # [seq_len, 1]
        self.positions = tf.cast(tf.range(seq_len)[:, None], tf.float32)

        # [1, model_dim // 2]
        angle_rates = 1 / (10000 ** self.depths)
        arc = (tf.multiply(self.positions, angle_rates))[None, ...]

        pe_sin = tf.sin(arc + phi)
        pe_cos = tf.cos(arc + phi)
        pe = tf.concat([pe_sin, pe_cos], axis=-1)

        return pe


class DNN(layers.Layer):
    """The Multi Layer Percetron
      Input shape
        - nD tensor with shape: ``(batch_size, ..., input_dim)``.
        The most common situation would be a 2D input with shape ``(batch_size, input_dim)``.
      Output shape
        - nD tensor with shape: ``(batch_size, ..., hidden_size[-1])``.
        For instance, for a 2D input with shape ``(batch_size, input_dim)``,
        the output would have shape ``(batch_size, hidden_size[-1])``.
      Arguments
        - **hidden_units**:list of positive integer, the layer number and units in each layer.
        - **activation**: Activation function to use.
        - **l2_reg**: float between 0 and 1. L2 regularizer strength applied to the kernel weights matrix.
        - **dropout_rate**: float in [0,1). Fraction of the units to dropout.
        - **use_bn**: bool. Whether use BatchNormalization before activation or not.
        - **output_activation**: Activation function to use in the last layer.If ``None``,
        it will be same as ``activation``.
        - **seed**: A Python integer to use as random seed.
    """

    def __init__(self, hidden_size, activation='relu', l2_reg=0, dropout_rate=0, use_bn=False,
                 output_activation=None,
                 **kwargs):
        self.hidden_size = hidden_size if isinstance(hidden_size, list) else [hidden_size]
        self.activation = activation
        self.l2_reg = l2_reg
        self.dropout_rate = dropout_rate
        self.use_bn = use_bn
        self.output_activation = output_activation

        self.dense_layers = [layers.Dense(self.hidden_size[i]) for i in range(len(self.hidden_size))]
        if self.use_bn:
            self.bn_layers = [tf.keras.layers.BatchNormalization() for _ in range(len(self.hidden_size))]

        self.dropout_layers = [tf.keras.layers.Dropout(self.dropout_rate) for _ in
                               range(len(self.hidden_size))]

        self.activation_layers = [activation_layer(self.activation) for _ in range(len(self.hidden_size))]

        if self.output_activation:
            self.activation_layers[-1] = activation_layer(self.output_activation)

        super(DNN, self).__init__(**kwargs)

    def call(self, inputs, training=None, **kwargs):

        deep_input = inputs

        for i in range(len(self.hidden_size)):
            fc = self.dense_layers[i](deep_input)

            if self.use_bn:
                fc = self.bn_layers[i](fc, training=training)

            fc = self.activation_layers[i](fc)

            fc = self.dropout_layers[i](fc, training=training)
            deep_input = fc

        return deep_input



import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow.keras import layers

from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
from easy_tpp.utils.tf_utils import get_shape_list

tfd = tfp.distributions
tfb = tfp.bijectors

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class MixtureSameFamily(tfd.MixtureSameFamily):
    """Mixture (same-family) distribution, redefined `log_cdf` and `log_survival_function`.
    """

    def log_cdf(self, x):
        x = x[..., None]
        log_cdf_x = self.components_distribution.log_cdf(x)
        mix_logits = self.mixture_distribution.logits
        return tf.reduce_logsumexp(log_cdf_x + mix_logits, axis=-1)

    def log_survival_function(self, x):
        x = x[..., None]
        log_sf_x = self.components_distribution.log_survival_function(x)
        mix_logits = self.mixture_distribution.logits
        return tf.reduce_logsumexp(log_sf_x + mix_logits, axis=-1)


class LogNormalMixtureDistribution:
    """
    Mixture of log-normal distributions.

    Args:
        locs (tensor): [batch_size, seq_len, num_mix_components].
        log_scales (tensor): [batch_size, seq_len, num_mix_components].
        log_weights (tensor): [batch_size, seq_len, num_mix_components].
        mean_log_inter_time (float): Average log-inter-event-time.
        std_log_inter_time (float): Std of log-inter-event-times.
    """

    def __init__(self, locs, log_scales, log_weights, mean_log_inter_time, std_log_inter_time, validate_args=None):
        mixture_dist = tfd.Categorical(logits=log_weights)
        component_dist = tfd.Normal(loc=locs, scale=tf.exp(log_scales))
        self.GMM = MixtureSameFamily(mixture_dist, component_dist)
        self.mean_log_inter_time = mean_log_inter_time
        self.std_log_inter_time = std_log_inter_time

        self.transformed_distribution = tfd.TransformedDistribution(self.GMM,
                                                                    bijector=tfb.Exp(),
                                                                    validate_args=validate_args)

    def log_prob(self, x):
        return self.transformed_distribution.log_prob(x)

    def log_survival_function(self, x):
        return self.transformed_distribution.log_survival_function(x)


class IntensityFree(TfBaseModel):
    """Tensorflow implementation of Intensity-Free Learning of Temporal Point Processes, ICLR 2020.
    https://openreview.net/pdf?id=HygOjhEYDH

    reference: https://github.com/shchur/ifl-tpp
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(IntensityFree, self).__init__(model_config)

        self.num_mix_components = model_config.data_specs['num_mix_components']
        self.num_features = 1 + self.hidden_size

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('IntensityFree'):
            self.build_input_graph()

            self.layer_rnn = layers.GRU(self.hidden_size,
                                        return_state=False,
                                        return_sequences=True)
            # activation='tanh')

            self.context_init = tf.zeros(self.hidden_size)[None, None, :]
            self.mark_linear = layers.Dense(self.num_event_types_pad)
            self.linear = layers.Dense(3 * self.num_mix_components)

            self.loss, self.num_event = self.loglike_loss()

            # Make predictions
            if self.event_sampler and self.gen_config.num_step_gen == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.event_sampler and self.gen_config.num_step_gen > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config.num_step_gen)

    def forward(self, time_delta_seqs, type_seqs):
        """Call the model.

        Args:
            time_delta_seqs (tensor): [batch_size, seq_len], inter-event time seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.

        Returns:
            tensor: hidden states, [batch_size, seq_len, hidden_dim], states right before the event happens.
        """
        # [batch_size, seq_len, 1]
        temporal_seqs = tf.log(time_delta_seqs + self.eps)[..., None]

        # [batch_size, seq_len, hidden_size]
        type_emb = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len, hidden_size + 1]
        features = tf.concat([temporal_seqs, type_emb], axis=-1)

        # [batch_size, seq_len, hidden_size]
        context = self.layer_rnn(features)

        batch_size, seq_len, hidden_size = get_shape_list(context)

        # (batch_size, 1, hidden_size)
        context_init = tf.tile(self.context_init, [batch_size, 1, 1])

        # (batch_size, seq_len + 1, hidden_size)
        context = tf.concat([context_init, context], axis=1)

        return context

    def loglike_loss(self):
        """Compute the loglike loss.

        Returns:
            tuple: loglikelihood loss and num of events.

        """

        time_delta_seqs = self.time_delta_seqs
        type_seqs = self.type_seqs
        batch_non_pad_mask = self.batch_non_pad_mask

        mean_log_inter_time = tf.reduce_mean(tf.log(time_delta_seqs))
        std_log_inter_time = tf.math.reduce_std(tf.log(time_delta_seqs))

        # [batch_size, seq_len, hidden_size]
        # seq_len = time_delta_seqs[:, 1:].size()[1]
        context = self.forward(time_delta_seqs[:, 1:], type_seqs[:, :-1])

        # (batch_size, seq_len, 3 * num_mix_components)
        raw_params = self.linear(context)
        locs = raw_params[..., :self.num_mix_components]
        log_scales = raw_params[..., self.num_mix_components: (2 * self.num_mix_components)]
        log_weights = raw_params[..., (2 * self.num_mix_components):]

        log_weights = tf.nn.log_softmax(log_weights, dim=-1)
        inter_time_dist = LogNormalMixtureDistribution(
            locs=locs,
            log_scales=log_scales,
            log_weights=log_weights,
            mean_log_inter_time=mean_log_inter_time,
            std_log_inter_time=std_log_inter_time
        )

        inter_times = tf.clip_by_value(time_delta_seqs, 1e-10, 1e10)
        # (batch_size, seq_len)
        log_p = inter_time_dist.log_prob(inter_times)

        # (batch_size, 1)
        # last_event_idx = tf.cast(tf.reduce_sum(batch_non_pad_mask, axis=-1, keepdims=True),
        #                          tf.int32) - 1

        log_surv_all = inter_time_dist.log_survival_function(inter_times)

        self.inter_times = log_surv_all

        #
        # # (batch_size,)
        # log_surv_last = tf.gather(log_surv_all, axis=-1, indices=last_event_idx)[..., None]

        # (batch_size, seq_len, num_marks)
        mark_logits = tf.nn.log_softmax(self.mark_linear(context), dim=-1)
        mark_dist = tfd.Categorical(logits=mark_logits)
        log_p += mark_dist.log_prob(type_seqs)

        # (batch_size, seq_len)
        log_p = tf.boolean_mask(log_p, batch_non_pad_mask) + self.eps
        # (batch_size,)
        loss = -tf.reduce_sum(log_p)

        num_events = get_shape_list(log_p)[0]

        return loss, num_events



import tensorflow as tf
from tensorflow.python.keras import layers

from easy_tpp.model.tf_model.tf_baselayer import EncoderLayer, TimePositionalEncoding
from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
from easy_tpp.utils.tf_utils import get_shape_list

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


class THP(TfBaseModel):
    """Tensorflow implementation of Transformer Hawkes Process, ICML 2020, https://arxiv.org/abs/2002.09291.
    """

    def __init__(self, model_config):
        """Intialiaze the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(THP, self).__init__(model_config)
        self.d_model = model_config.hidden_size
        self.d_time = model_config.time_emb_size
        self.use_norm = model_config.use_ln

        self.n_layers = model_config.num_layers
        self.n_head = model_config.num_heads
        self.dropout = model_config.dropout_rate

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('THP'):
            self.build_input_graph()

            self.layer_temporal_encoding = TimePositionalEncoding(self.d_model)

            self.factor_intensity_base = tf.get_variable(name='intensity_base', shape=[1, self.num_event_types])
            self.factor_intensity_decay = tf.get_variable(name='intensity_decay',
                                                          shape=[1, self.num_event_types])

            # convert hidden vectors into event-type-sized vector
            self.layer_intensity_hidden = layers.Dense(self.num_event_types)
            self.softplus = tf.nn.softplus

            self.layer_intensity = layers.Dense(self.num_event_types, activation=tf.nn.softplus)

            self.stack_layers = [EncoderLayer(hidden_size=self.d_model,
                                              num_heads=self.n_head,
                                              dropout_rate=self.dropout) for _ in range(self.n_layers)]

            self.loss, self.num_event = self.loglike_loss()

            # Make predictions
            if self.gen_config and self.gen_config['num_step_gen'] == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.gen_config and self.gen_config['num_step_gen'] > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config.num_step_gen)

    def forward(self, time_seqs, type_seqs, attention_mask):
        """Call the model

        Args:
            time_seqs (tensor): [batch_size, seq_len], timestamp seqs.
            time_delta_seqs (tensor): [batch_size, seq_len], inter-event time seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.
            attention_mask (tensor): [batch_size, seq_len, hidden_size], attention masks.

        Returns:
            tensor: hidden states at event times.
        """
        # [batch_size, seq_len, hidden_size]
        tem_enc = self.layer_temporal_encoding(time_seqs)
        enc_output = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len, hidden_size]
        for enc_layer in self.stack_layers:
            enc_output += tem_enc
            enc_output = enc_layer((
                enc_output,
                attention_mask))
        return enc_output

    def loglike_loss(self):
        """Compute the loglike loss.

        Args:
            batch (list): batch input.

        Returns:
            list: loglike loss, num events.
        """
        # 1. compute event-loglik
        enc_out = self.forward(self.time_seqs[:, :-1],
                               self.type_seqs[:, :-1],
                               self.attention_mask[:, 1:, :-1])

        # [1, 1, num_event_types]
        factor_intensity_decay = self.factor_intensity_decay[None, ...]
        factor_intensity_base = self.factor_intensity_base[None, ...]

        # update time decay based on Equation (6)
        # [batch_size, seq_len, num_event_types]
        intensity_states = factor_intensity_decay * self.time_delta_seqs[:, 1:, None] + self.layer_intensity_hidden(
            enc_out) + factor_intensity_base

        lambda_at_event = self.softplus(intensity_states)

        # 2. compute non-event-loglik (using MC sampling to compute integral)
        # 2.1 sample times
        # [batch_size, seq_len, num_sample]
        sample_dtimes = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # 2.2 compute intensities at sampled times
        # [batch_size, num_times = max_len - 1, num_sample, event_num]
        state_t_sample = self.compute_states_at_sample_times(event_states=enc_out,
                                                             sample_dtimes=sample_dtimes)
        lambda_t_sample = self.softplus(state_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        # return enc_inten to compute accuracy
        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_states_at_sample_times(self, event_states, sample_dtimes):
        """Compute the hidden states at sampled times.

        Args:
            event_states (tensor): [batch_size, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].

        Returns:
            tensor: hidden state at each sampled time.
        """
        # [batch_size, seq_len, 1, hidden_size]
        event_states = event_states[:, :, None, :]

        # [batch_size, seq_len, num_samples, 1]
        sample_dtimes = sample_dtimes[..., None]

        # [1, 1, 1, num_event_types]
        factor_intensity_decay = self.factor_intensity_decay[None, None, ...]
        factor_intensity_base = self.factor_intensity_base[None, None, ...]

        # update time decay based on Equation (6)
        # [batch_size, seq_len, num_samples, num_event_types]
        intensity_states = factor_intensity_decay * sample_dtimes + self.layer_intensity_hidden(
            event_states) + factor_intensity_base

        return intensity_states

    def compute_intensities_at_sample_times(self,
                                            time_seqs,
                                            time_delta_seqs,
                                            type_seqs,
                                            sample_dtimes,
                                            **kwargs):
        """Compute the intensity at sampled times.

         Args:
             time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
             time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
             type_seqs (tensor): [batch_size, seq_len], sequences of event types.
             sampled_dtimes (tensor): [batch_size, seq_len, num_sample], sampled time delta sequence.

         Returns:
             tensor: intensities as sampled_dtimes, [batch_size, seq_len, num_samples, event_num].
         """

        attention_mask = kwargs.get('attention_mask', None)
        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        if attention_mask is None:
            batch_size, seq_len = get_shape_list(time_seqs)
            attention_mask = tf.ones((seq_len, seq_len))
            # only keep the strict upper triangular
            lower_diag_masks = tf.linalg.LinearOperatorLowerTriangular(tf.ones_like(attention_mask)).to_dense()
            attention_mask = tf.where(tf.equal(lower_diag_masks, 0),
                                      attention_mask,
                                      tf.zeros_like(attention_mask))
            attention_mask = tf.tile(attention_mask[None, ...], (batch_size, 1, 1))
            attention_mask = tf.cast(attention_mask, tf.int32)

        # [batch_size, seq_len, num_samples]
        enc_out = self.forward(time_seqs, type_seqs, attention_mask)

        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = self.compute_states_at_sample_times(enc_out, sample_dtimes)

        if compute_last_step_only:
            lambdas = self.layer_intensity(encoder_output[:, -1:, :, :])
        else:
            # [batch_size, seq_len, num_samples, num_event_types]
            lambdas = self.layer_intensity(encoder_output)
        return lambdas



import numpy as np
import tensorflow as tf
from tensorflow.keras import layers

from easy_tpp.model.tf_model.tf_baselayer import DNN
from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
from easy_tpp.utils import rk4_step_method
from easy_tpp.utils.tf_utils import get_shape_list

if tf.__version__ >= '2.0':
    tf = tf.compat.v1
    tf.disable_v2_behavior()


def get_neural_ode_layer(
        ode_fn,
        num_samples=10,
        solver=rk4_step_method,
        return_states=False
):
    """Get a black-box neural ode layer parameterized by parameters.

    Args:
        ode_fn: function
            It likes f(solver_function, dt, z_list), and contains the learnable variables.
        num_samples: int
            Number of samples in time interval dt.
        solver: function
            Solver function like f(ode_func, dt, z_0)
        return_states: bool, default False
            Identify whether return whole states or just last state.

    Returns:
        A neural_ode_layer (function) with signature f(z0, dt).
    """

    @tf.custom_gradient
    def neural_ode_layer(
            z0,
            dt,
    ):
        """Calculate z1 by z0 and time gap dt.

        Args:
            z0: Tensor with shape [..., dim]
            dt: Tensor with shape [..., 1 or dim]

        Returns:
            A tensor presents z1, whose shape is the same as z0.
        """
        with tf.name_scope('neural_ode'):
            # Forward activity
            dt_ratio = 1.0 / num_samples
            delta_t = dt * dt_ratio

            z = z0
            z_list = []
            for i in range(num_samples):
                z = solver(ode_fn, delta_t, z)
                z_list.append(z)
            z1 = z

            def grad(a1, variables=None):
                # a1 is grad_z1 == dL/dz1
                if variables is None:
                    variables = []

                def aug_dynamics(tmp_states):
                    """
                    Ode function for states [z_1, a_1, \thetas (many)].

                    Args:
                        tmp_states: list
                            Elements are [z_1, a_1, \thetas (many)].

                    Returns:
                        List contains differentiations of states.
                    """

                    tmp_z = tmp_states[0]
                    tmp_neg_a = -tmp_states[1]
                    # tmp_var_grad = tmp_states[2:]

                    # calculate dz/dt

                    # if tf.__version__ < '2.0':
                    #     # using GradientType to calculate (faster when building graph)
                    #     with tf.GradientTape() as g:
                    #         g.watch([tmp_z, *variables])
                    #         func_eval = ode_fn(tmp_z)
                    #         tmp_ds = g.gradient(func_eval, [tmp_z, *variables], output_gradients=tmp_neg_a)
                    # else:
                    # using tf.gradients to calculate
                    func_eval = ode_fn(tmp_z)
                    tmp_ds = tf.gradients(func_eval, [tmp_z, *variables], grad_ys=tmp_neg_a)

                    neg_adfdz = tmp_ds[0]
                    neg_adfdtheta = [tf.reshape(var, [-1]) for var in tmp_ds[1:]]

                    return [func_eval, neg_adfdz, *neg_adfdtheta]

                # Backward activity
                if tf.__version__ < '2.0':
                    # Compile EAGER graph to static (this will be much faster)
                    import tensorflow.contrib.eager as tfe
                    aug_dynamics = tfe.defun(aug_dynamics)

                # Construct back-state for ode solver
                # reshape variable \theta for batch solving
                init_var_grad = [tf.zeros([np.prod(get_shape_list(var))]) for var in variables]

                if a1 is None:
                    a1 = tf.zeros_like(z1)

                # [z(t_1), a(t_1), \theta]
                states = [z1, a1, *init_var_grad]
                # print('states:', states)
                for i in range(num_samples):
                    states = solver(aug_dynamics, -delta_t, states)

                grad_z0 = states[1]
                grad_t = tf.ones_like(dt)

                if variables is not None:
                    # average the different dt effect on variable \theta
                    grad_theta = [tf.reshape(tf.reduce_mean(var_grad, axis=0), var.shape) for var, var_grad in
                                  zip(variables, states[2:])]
                    return (grad_z0, grad_t), grad_theta
                else:
                    return grad_z0, grad_t

        if return_states:
            return z_list, grad
        else:
            return z1, grad

    return neural_ode_layer


class ODETPP(TfBaseModel):
    """
    TODO: this version runs ok under tf 1.13, but is very slow under tf 2.0
    """
    def __init__(self, model_config):
        super(ODETPP, self).__init__(model_config)
        self.ode_num_sample_per_step = model_config.model_specs['ode_num_sample_per_step']
        self.time_factor = model_config.model_specs['time_factor']
        self.seq_len = model_config.max_len

    def build_graph(self):
        """Build up the network
        """
        with tf.variable_scope('ODETPP'):
            # have to specify the max len of the input to avoid a variable length of tensor.
            # for looping over the variable length of tensor, custom gradient can not properly work
            # in the scan (while_loop)
            # tf.GradientTape.gradients() does not support graph control flow operations
            # like tf.cond or tf.while at this time

            # Input placeholder
            # shape - (batch_size, max_len)
            # max_len - sequence length including time zero padding
            self.time_delta_seqs = tf.placeholder(tf.float32, shape=[None, self.seq_len])

            # shape - (batch_size, max_len)
            self.time_seqs = tf.placeholder(tf.float32, shape=[None, self.seq_len])

            # shape - (batch_size, max_len)
            self.type_seqs = tf.placeholder(tf.int32, shape=[None, self.seq_len])

            # shape - (batch_size, max_len)
            self.batch_non_pad_mask = tf.placeholder(tf.int32, shape=[None, self.seq_len])

            # shape - (batch_size, max_len, max_len)
            self.attention_mask = tf.placeholder(tf.int32, shape=[None, None, None])

            # Event type one-hot code
            # shape - (batch_size, max_len, num_event_types)
            self.type_mask = tf.placeholder(tf.float32, shape=[None, None, None])

            self.layer_intensity = layers.Dense(self.num_event_types, activation=tf.nn.softplus)

            self.event_model = DNN(hidden_size=self.hidden_size)

            self.solver = rk4_step_method

            self.layer_neural_ode = get_neural_ode_layer(ode_fn=self.event_model,
                                                         solver=self.solver,
                                                         num_samples=self.ode_num_sample_per_step)

            self.loss, self.num_event = self.loglike_loss()

            self.is_training = tf.placeholder(tf.bool)

            # Make predictions
            if self.event_sampler and self.gen_config.num_step_gen == 1:
                self.dtime_predict_one_step, self.type_predict_one_step = \
                    self.predict_one_step_at_every_event(self.time_seqs,
                                                         self.time_delta_seqs,
                                                         self.type_seqs)

            if self.event_sampler and self.gen_config.num_step_gen > 1:
                # make generations
                self.dtime_generation, self.type_generation = \
                    self.predict_multi_step_since_last_event(self.time_seqs,
                                                             self.time_delta_seqs,
                                                             self.type_seqs,
                                                             num_step=self.gen_config.num_step_gen)

    def forward(self, time_delta_seqs, type_seqs, **kwargs):
        """Call the model.

        Args:
            time_delta_seqs (tensor): [batch_size, seq_len], inter-event time seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.

        Returns:
            tensor: hidden states at event times.

        """
        # [batch_size, seq_len=max_len, hidden_size]
        type_seq_emb = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len=max_len, 1]
        time_delta_seqs_ = time_delta_seqs[..., None]

        total_state_at_event_minus = []
        total_state_at_event_plus = []
        last_state = tf.zeros_like(type_seq_emb[:, 0, :])
        for type_emb, dt in zip(tf.unstack(type_seq_emb, axis=-2),
                                tf.unstack(time_delta_seqs_, axis=-2)):
            # the bp may break for tf 1.13 when dt is large
            # after testing, we put a time factor here to avoid the failure of bp
            # it is not needed for tf 2.0.
            dt = dt / self.time_factor
            last_state = self.layer_neural_ode(last_state + type_emb, dt)
            total_state_at_event_minus.append(last_state)
            total_state_at_event_plus.append(last_state + type_emb)

        # [batch_size, seq_len, hidden_size]
        state_ti = tf.stack(total_state_at_event_minus, axis=1)

        # [batch_size, seq_len, hidden_size]
        state_to_evolve = tf.stack(total_state_at_event_plus, axis=1)

        return state_ti, state_to_evolve

    def loglike_loss(self):
        """Compute the loglike loss.

        Args:
            batch (list): batch input.

        Returns:
            list: loglike loss, num events.
        """

        state_ti, state_ti_plus = self.forward(self.time_delta_seqs[:, 1:], self.type_seqs[:, :-1])

        # Lambda(t) right before each event time point
        # lambda_at_event - [batch_size, num_times=max_len-1, num_event_types]
        # Here we drop the last event because it has no delta_time label (can not decay)
        lambda_at_event = self.layer_intensity(state_ti)

        # interval_t_sample - [batch_size, num_times=max_len-1, num_mc_sample]
        # for every batch and every event point => do a sampling (num_mc_sampling)
        # the first dtime is zero, so we use time_delta_seq[:, 1:]
        interval_t_sample = self.make_dtime_loss_samples(self.time_delta_seqs[:, 1:])

        # [batch_size, num_times = max_len - 1, num_mc_sample, hidden_size]
        sample_state_ti = self.compute_states_at_sample_times(state_ti_plus, interval_t_sample)

        # [batch_size, num_times = max_len - 1, num_mc_sample, event_num]
        lambda_t_sample = self.layer_intensity(sample_state_ti)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=self.time_delta_seqs[:, 1:],
                                                                        seq_mask=self.batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=self.type_mask[:, 1:])

        loss = - tf.reduce_sum(event_ll - non_event_ll)

        return loss, num_events

    def compute_states_at_sample_times(self, state_ti_plus, sample_dtimes):
        """Compute the states at sampling times.

        Args:
            state_ti_plus (tensor): [batch_size, seq_len, hidden_size], states right after the events.
            sample_dtimes (tensor): [batch_size, seq_len, num_samples], delta times in sampling.

        Returns:
            tensor: hiddens states at sampling times.
        """

        # Use broadcasting to compute the decays at all time steps
        # at all sample points
        # h_ts shape (batch_size, seq_len, num_samples, hidden_dim)
        state = self.solver(diff_func=self.event_model,
                            dt=sample_dtimes[..., None],  # [batch_size, seq_len, num_samples, 1]
                            z0=state_ti_plus[..., None, :])  # [batch_size, seq_len, 1, hidden_size]

        return state

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_dtimes, **kwargs):
        """Compute the intensity at sampled times, not only event times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], times seqs.
            time_delta_seqs (tensor): [batch_size, seq_len], time delta seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.
            sample_dtimes (tensor): [batch_size, seq_len, num_sample], sampled inter-event timestamps.

        Returns:
            tensor: [batch_size, num_times, num_mc_sample, num_event_types],
                    intensity at each timestamp for each event type.
        """

        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        state_ti, state_ti_plus = self.forward(time_delta_seqs, type_seqs, **kwargs)

        # Num of samples in each batch and num of event time point in the sequence
        batch_size, seq_len, _ = get_shape_list(state_ti)

        if compute_last_step_only:
            interval_t_sample = sample_dtimes[:, -1:, :]
        else:
            # interval_t_sample - [batch_size, num_times, num_mc_sample, 1]
            interval_t_sample = sample_dtimes
            # Use broadcasting to compute the decays at all time steps
            # at all sample points

        # [batch_size, num_sample_times / 1, num_mc_sample, hidden_size]
        sample_state_ti = self.compute_states_at_sample_times(state_ti_plus, interval_t_sample)

        # [batch_size, num_sample_times / 1, num_mc_sample, num_event_types]
        sampled_intensities = self.layer_intensity(sample_state_ti)

        return sampled_intensities



from easy_tpp.model.torch_model.torch_anhn import ANHN as TorchANHN
from easy_tpp.model.torch_model.torch_attnhp import AttNHP as TorchAttNHP
from easy_tpp.model.torch_model.torch_basemodel import TorchBaseModel
from easy_tpp.model.torch_model.torch_fullynn import FullyNN as TorchFullyNN
from easy_tpp.model.torch_model.torch_intensity_free import IntensityFree as TorchIntensityFree
from easy_tpp.model.torch_model.torch_nhp import NHP as TorchNHP
from easy_tpp.model.torch_model.torch_ode_tpp import ODETPP as TorchODETPP
from easy_tpp.model.torch_model.torch_rmtpp import RMTPP as TorchRMTPP
from easy_tpp.model.torch_model.torch_sahp import SAHP as TorchSAHP
from easy_tpp.model.torch_model.torch_thp import THP as TorchTHP

# by default, we use torch and do not install tf, therefore we ignore the import error
try:
    from easy_tpp.model.tf_model.tf_basemodel import TfBaseModel
    from easy_tpp.model.tf_model.tf_nhp import NHP as TfNHP
    from easy_tpp.model.tf_model.tf_ode_tpp import ODETPP as TfODETPP
    from easy_tpp.model.tf_model.tf_thp import THP as TfTHP
    from easy_tpp.model.tf_model.tf_sahp import SAHP as TfSAHP
    from easy_tpp.model.tf_model.tf_rmtpp import RMTPP as TfRMTPP
    from easy_tpp.model.tf_model.tf_attnhp import AttNHP as TfAttNHP
    from easy_tpp.model.tf_model.tf_anhn import ANHN as TfANHN
    from easy_tpp.model.tf_model.tf_fullynn import FullyNN as TfFullyNN
    from easy_tpp.model.tf_model.tf_intensity_free import IntensityFree as TfIntensityFree
except ImportError:
    pass

__all__ = ['TorchBaseModel',
           'TorchNHP',
           'TorchAttNHP',
           'TorchTHP',
           'TorchSAHP',
           'TorchFullyNN',
           'TorchIntensityFree',
           'TorchODETPP',
           'TfBaseModel',
           'TfNHP',
           'TfAttNHP',
           'TfTHP',
           'TfSAHP',
           'TfANHN',
           'TfFullyNN',
           'TfIntensityFree',
           'TfODETPP']



import codecs
import os
import re
from setuptools import find_packages
from setuptools import setup


def readme():
    with codecs.open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content


def get_version():
    version_file = os.path.join(os.path.dirname(__file__), "version.py")
    version_regex = r"__version__ = ['\"]([^'\"]*)['\"]"
    with open(version_file, "r") as f:
        version = re.search(version_regex, f.read(), re.M).group(1)
        return version


def parse_requirements(fname='requirements.txt'):
    """Parse the package dependencies listed in a requirements file."""

    def parse_line(line):
        """Parse information from a line in a requirements text file."""
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            for line in parse_require_file(target):
                yield line
        else:
            yield line

    def parse_require_file(fpath):
        with codecs.open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for ll in parse_line(line):
                        yield ll

    packages = list(parse_require_file(fname))
    return packages


setup(
    name='easy_tpp',
    version=get_version(),
    description='An easy and flexible tool for neural temporal point process',
    url = 'https://github.com/ant-research/EasyTemporalPointProcess/',
    # long_description = 'Our EasyTPP makes several unique contributions to this area: a unified interface of using existing datasets and adding new datasets; a wide range of evaluation programs that are easy to use and extend as well as facilitate reproducible research; implementations of popular neural TPPs, together with a rich library of modules by composing which one could quickly build complex models. ',
    # long_description=open('README.md').read(),
    # long_description_content_type='text/markdown',
    author='Alipay',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    install_requires=parse_requirements('requirements.txt'))



def ode_update_op(z0, dz, dt):
    """
    General update operation for solving ODEs.

    Args:
        z0: Tensor or a list for Tensor whose shape is [..., dim]
            State at t0.
        dz: Tensor or a list for Tensor whose shape is [..., dim]
            Differentiation of state.
        dt: Tensor with shape [..., 1]
            Equal to t1 - t0.

    Returns:

    """
    if isinstance(z0, list) or isinstance(z0, tuple):
        return [item_z + dt * item_dz for item_z, item_dz in zip(z0, dz)]
    else:
        return z0 + dt * dz


def euler_step_method(diff_func, dt, z0):
    """
    Euler method for solving ODEs.

    Args:
        diff_func: function(state)
            Differential equation.
        dt: Tensor with shape [..., 1]
            Equal to t1 - t0.
        z0: Tensor or a list for Tensor whose shape is [..., dim]
            State at t0.

    Returns:
        Tensor or a list for Tensor whose shape is [..., dim], which is updated state.
    """
    dz = diff_func(z0)
    return ode_update_op(z0, dz, dt)


def rk2_step_method(diff_func, dt, z0):
    """
    Second order Runge-Kutta method for solving ODEs.

    Args:
        diff_func: function(dt, state)
            Differential equation.
        dt: Tensor with shape [..., 1]
            Equal to t1 - t0.
        z0: Tensor or a list for Tensor whose shape is [..., dim]
            State at t0.

    Returns:
        Tensor or a list for Tensor whose shape is [..., dim]
    """
    # shape -> [..., dim]
    k1 = diff_func(z0)
    k2 = diff_func(ode_update_op(z0, k1, dt))

    if isinstance(z0, list) or isinstance(z0, tuple):
        return [item_z + (item_k1 + item_k2) * dt * 0.5 for item_z, item_k1, item_k2 in zip(z0, k1, k2)]
    else:
        return z0 + dt * (k1 + k2) * 0.5


def rk4_step_method(diff_func, dt, z0):
    """
    Fourth order Runge-Kutta method for solving ODEs.

    Args:
        diff_func: function(dt, state)
            Differential equation.
        dt: Tensor with shape [..., 1]
            Equal to t1 - t0.
        z0: Tensor with shape [..., dim]
            State at t0.

    Returns:
        Tensor with shape [..., dim], which is updated state.
    """
    # shape -> [..., dim]
    k1 = diff_func(z0)
    k2 = diff_func(ode_update_op(z0, k1, dt / 2.0))
    k3 = diff_func(ode_update_op(z0, k2, dt / 2.0))
    k4 = diff_func(ode_update_op(z0, k3, dt))

    if isinstance(z0, list) or isinstance(z0, tuple):
        return [item_z + (item_k1 + 2.0 * item_k2 + 2.0 * item_k3 + item_k4) * dt / 6.0
                for item_z, item_k1, item_k2, item_k3, item_k4 in zip(z0, k1, k2, k3, k4)]
    else:
        return z0 + dt * (k1 + k2 * 2.0 + k3 * 2.0 + k4) / 6.0



import torch
from torch import nn

from easy_tpp.model.torch_model.torch_basemodel import TorchBaseModel


class RMTPP(TorchBaseModel):
    """Torch implementation of Recurrent Marked Temporal Point Processes, KDD 2016.
    https://www.kdd.org/kdd2016/papers/files/rpp1081-duA.pdf
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(RMTPP, self).__init__(model_config)

        self.layer_temporal_emb = nn.Linear(1, self.hidden_size)

        self.layer_rnn = nn.RNN(input_size=self.hidden_size, hidden_size=self.hidden_size,
                                num_layers=1, batch_first=True)

        self.layer_hidden = nn.Linear(self.hidden_size, self.num_event_types)

        self.factor_intensity_base = torch.empty([1, 1, self.num_event_types], device=self.device)
        self.factor_intensity_current_influence = torch.empty([1, 1, self.num_event_types], device=self.device)

        nn.init.xavier_normal_(self.factor_intensity_base)
        nn.init.xavier_normal_(self.factor_intensity_current_influence)

    def state_decay(self, states_to_decay, duration_t):
        """Equation (11), which computes intensity
        """

        # [batch_size, seq_len, num_event_types]
        states_to_decay_ = self.layer_hidden(states_to_decay)

        # [batch_size, seq_len, num_event_types]
        # put a max number to avoid explode during HPO
        intensity = torch.exp(
            states_to_decay_ + self.factor_intensity_current_influence * duration_t +
            self.factor_intensity_base).clamp(max=1e5)
        return intensity

    def forward(self, time_seqs, time_delta_seqs, type_seqs, **kwargs):
        """Call the model.

        Args:
            batch (list): batch input.

        Returns:
            list: hidden states, [batch_size, seq_len, hidden_dim], states right before the event happens;
                  stacked decay states,  [batch_size, max_seq_length, 4, hidden_dim], states right after
                  the event happens.
        """

        max_steps = kwargs.get('max_steps', None)

        # last event has no time label
        max_seq_length = max_steps if max_steps is not None else type_seqs.size(1) - 1

        # [batch_size, seq_len, hidden_size]
        type_emb = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len, hidden_size]
        temporal_emb = self.layer_temporal_emb(time_seqs[..., None])

        # [batch_size, seq_len, hidden_size]
        # states right after the event
        decay_states, _ = self.layer_rnn(type_emb + temporal_emb)

        # if only one event, then we dont decay
        if max_seq_length == 1:
            h_t = decay_states
        else:
            # States decay - Equation (7) in the paper
            # states before the happening of the next event
            h_t = self.state_decay(decay_states, time_delta_seqs[..., None])

        return h_t, decay_states

    def loglike_loss(self, batch):
        """Compute the loglike loss.

        Args:
            batch (list): batch input.

        Returns:
            tuple: loglikelihood loss and num of events.
        """
        time_seqs, time_delta_seqs, type_seqs, batch_non_pad_mask, _, type_mask = batch

        lambda_at_event, decay_states = self.forward(time_seqs[:, :-1], time_delta_seqs[:, 1:], type_seqs[:, :-1])

        # Num of samples in each batch and num of event time point in the sequence
        batch_size, seq_len, _ = lambda_at_event.size()

        # Compute the big lambda integral in equation (8)
        # 1 - take num_mc_sample rand points in each event interval
        # 2 - compute its lambda value for every sample point
        # 3 - take average of these sample points
        # 4 - times the interval length

        # interval_t_sample - [batch_size, num_times=max_len-1, num_mc_sample]
        # for every batch and every event point => do a sampling (num_mc_sampling)
        # the first dtime is zero, so we use time_delta_seqs[:, 1:]
        interval_t_sample = self.make_dtime_loss_samples(time_delta_seqs[:, 1:])

        # [batch_size, num_times = max_len - 1, num_sample, event_num]
        lambda_t_sample = self.compute_states_at_sample_times(decay_states, interval_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=time_delta_seqs[:, 1:],
                                                                        seq_mask=batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=type_mask[:, 1:])

        # (num_samples, num_times)
        loss = - (event_ll - non_event_ll).sum()
        return loss, num_events

    def compute_states_at_sample_times(self, decay_states, sample_dtimes):
        """Compute the hidden states at sampled times.

        Args:
            decay_states (tensor): [batch_size, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].

        Returns:
            tensor: hidden state at each sampled time.
        """
        # update the states given last event

        # Use broadcasting to compute the decays at all time steps
        # decay_states[..., None, :]: [batch_size, seq_len, 1, hidden_size]
        # sample_dtimes[..., None]: [batch_size, seq_len, num_mc_sample, 1]
        # h_ts shape (batch_size, num_times, num_mc_sample, hidden_dim)
        h_ts = self.state_decay(decay_states[..., None, :],
                                sample_dtimes[..., None])

        return h_ts

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_times, **kwargs):
        """Compute the intensity at sampled times, not only event times.

        Args:
            time_seq (tensor): [batch_size, seq_len], times seqs.
            time_delta_seq (tensor): [batch_size, seq_len], time delta seqs.
            event_seq (tensor): [batch_size, seq_len], event type seqs.
            sample_dtimes (tensor): [batch_size, seq_len, num_sample], sampled inter-event timestamps.

        Returns:
            tensor: [batch_size, num_times, num_mc_sample, num_event_types],
                    intensity at each timestamp for each event type.
        """

        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        # forward to the last but one event
        _, decay_states = self.forward(time_seqs, time_delta_seqs, type_seqs, **kwargs)

        # Num of samples in each batch and num of event time point in the sequence
        batch_size, seq_len, _ = decay_states.size()

        if compute_last_step_only:
            interval_t_sample = sample_times[:, -1:, :, None]
            # [batch_size, 1, num_mc_sample, num_event_types]
            sampled_intensities = self.state_decay(decay_states[:, -1:, None, :],
                                                   interval_t_sample)

        else:
            # interval_t_sample - [batch_size, num_times, num_mc_sample, 1]
            interval_t_sample = sample_times[..., None]
            # Use broadcasting to compute the decays at all time steps
            # at all sample points
            # sampled_intensities shape (batch_size, num_times, num_mc_sample, hidden_dim)
            # decay_states[:, :, None, :]  (batch_size, num_times, 1, hidden_dim)
            sampled_intensities = self.state_decay(decay_states[..., None, :],
                                                   interval_t_sample)

        return sampled_intensities



from abc import abstractmethod
from collections import defaultdict
from typing import List

from easy_tpp.utils import logger, Registrable


class HyperTuner(Registrable):
    _trial_register_center = defaultdict(dict)

    def __init__(self, config, trial_end_callbacks: List[callable] = None):
        """Initialize the tuner

        Args:
            config (EasyTPP.Config): config class
            trial_end_callbacks (List[callable]): List of callback functions to be executed after each trial.
        """
        self.config = config
        self.trial_end_callbacks = trial_end_callbacks or []
        logger.info(f'Storage of hpo framework: {self.config.hpo_config.storage_uri}')

    @abstractmethod
    def get_all_best_runner_configs(self):
        pass

    @abstractmethod
    def get_best_runner_config_by_name(self, runner_id):
        """

        Args:
            runner_id (str):

        Returns:

        """
        pass

    @abstractmethod
    def get_num_remain_trials_by_name(self, runner_id):
        pass

    @staticmethod
    def build_from_config(config, trial_end_callbacks: List[callable] = None):
        """Load yaml config file from disk.

        Args:
            config (EasyTPP.Config): config class
            trial_end_callbacks (List[callable]): List of callback functions to be executed after each trial.

        Returns:
            EasyTPP.Config: Config object corresponding to cls.
        """
        runner_cls = HyperTuner.by_name(config.hpo_config.framework_id)
        return runner_cls(config, trial_end_callbacks)

    # ---------------------- Trail Register and Get Functions ---------------------

    @classmethod
    def register_trial_func(cls, model_id, overwrite=True):
        """Register the trial functions in HPO

        Args:
            model_id (str): id of the models.
            overwrite (bool, optional): whether to overwrite the trial function. Defaults to True.

        Returns:
            dict: the registered trial function
        """
        register_center = HyperTuner._trial_register_center

        def _register_trial(func):
            if model_id in register_center[cls]:
                if overwrite:
                    register_center[cls][model_id] = func
                    logger.info(f'The trial for {model_id} is already registered, but overwrite it.')
                else:
                    logger.warn(f'The trial for {model_id} is already registered, and cannot be overwritten!')
            else:
                register_center[cls][model_id] = func
                logger.info(f'Trial register: {cls.get_registered_name()} - {model_id}')
            return func

        return _register_trial

    @classmethod
    def retrieve_trial_func_by_model_name(cls, name):
        """Retrieve the trail function by the model id

        Args:
            name (str): model id.

        Raises:
            RuntimeError: non registered error for the hpo framework.

        Returns:
            dict: registered trial center
        """
        cls_trial_rc = HyperTuner._trial_register_center[cls]
        if name not in cls_trial_rc:
            if 'default' in cls_trial_rc:
                logger.warn(
                    f'Trial for {name} in {cls.get_registered_name()} is not existed, and use default trial!'
                )
                name = 'default'
            else:
                raise RuntimeError(f'This HPO Framework is not registered!')
        return cls_trial_rc[name]

    @classmethod
    def get_registered_name(cls):
        """Get the name of the registered hpo class.

        Returns:
            str: the name of the registered hpo class.
        """
        hpo_rc = HyperTuner.registry_dict()
        for registered_name, hpo_cls in hpo_rc.items():
            if cls in hpo_cls:
                return registered_name

        logger.warn(f'The hpo framework is not registered: {cls}')
        return None

    @abstractmethod
    def run(self):
        """Run the process.

        Raises:
            NotImplementedError: error raised in base class.
        """
        raise NotImplementedError



import torch
import torch.nn as nn

from easy_tpp.model.torch_model.torch_baselayer import EncoderLayer, MultiHeadAttention, TimePositionalEncoding
from easy_tpp.model.torch_model.torch_basemodel import TorchBaseModel


class THP(TorchBaseModel):
    """Torch implementation of Transformer Hawkes Process, ICML 2020, https://arxiv.org/abs/2002.09291.
    Note: Part of the code is collected from https://github.com/yangalan123/anhp-andtt/tree/master/thp.
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(THP, self).__init__(model_config)
        self.d_model = model_config.hidden_size
        self.d_time = model_config.time_emb_size
        self.use_norm = model_config.use_ln

        self.n_layers = model_config.num_layers
        self.n_head = model_config.num_heads
        self.dropout = model_config.dropout_rate

        self.layer_temporal_encoding = TimePositionalEncoding(self.d_model, device=self.device)

        self.factor_intensity_base = torch.empty([1, self.num_event_types], device=self.device)
        self.factor_intensity_decay = torch.empty([1, self.num_event_types], device=self.device)
        nn.init.xavier_normal_(self.factor_intensity_base)
        nn.init.xavier_normal_(self.factor_intensity_decay)

        # convert hidden vectors into event-type-sized vector
        self.layer_intensity_hidden = nn.Linear(self.d_model, self.num_event_types)
        self.softplus = nn.Softplus()

        self.stack_layers = nn.ModuleList(
            [EncoderLayer(
                self.d_model,
                MultiHeadAttention(self.n_head, self.d_model, self.d_model, self.dropout,
                                   output_linear=False),

                use_residual=False,
                dropout=self.dropout
            ) for _ in range(self.n_layers)])

    def forward(self, time_seqs, type_seqs, attention_mask):
        """Call the model

        Args:
            time_seqs (tensor): [batch_size, seq_len], timestamp seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.
            attention_mask (tensor): [batch_size, seq_len, hidden_size], attention masks.

        Returns:
            tensor: hidden states at event times.
        """
        # [batch_size, seq_len, hidden_size]
        tem_enc = self.layer_temporal_encoding(time_seqs)
        enc_output = self.layer_type_emb(type_seqs)

        # [batch_size, seq_len, hidden_size]
        for enc_layer in self.stack_layers:
            enc_output += tem_enc
            enc_output = enc_layer(
                enc_output,
                mask=attention_mask)

        return enc_output

    def loglike_loss(self, batch):
        """Compute the loglike loss.

        Args:
            batch (tuple, list): batch input.

        Returns:
            tuple: loglike loss, num events.
        """
        time_seqs, time_delta_seqs, type_seqs, batch_non_pad_mask, attention_mask, type_mask = batch

        # 1. compute event-loglik
        # [batch_size, seq_len, hidden_size]
        enc_out = self.forward(time_seqs[:, :-1], type_seqs[:, :-1], attention_mask[:, 1:, :-1])

        # [batch_size, seq_len, num_event_types]
        # update time decay based on Equation (6)
        # [1, 1, num_event_types]
        factor_intensity_decay = self.factor_intensity_decay[None, ...]
        factor_intensity_base = self.factor_intensity_base[None, ...]

        # update time decay based on Equation (6)
        # [batch_size, seq_len, num_event_types]
        intensity_states = factor_intensity_decay * time_delta_seqs[:, 1:, None] + self.layer_intensity_hidden(
            enc_out) + factor_intensity_base

        lambda_at_event = self.softplus(intensity_states)

        # 2. compute non-event-loglik (using MC sampling to compute integral)
        # 2.1 sample dtimes
        # [batch_size, seq_len, num_sample]
        sample_dtimes = self.make_dtime_loss_samples(time_delta_seqs[:, 1:])

        # 2.2 compute intensities at sampled times
        # [batch_size, num_times = max_len - 1, num_sample, event_num]
        state_t_sample = self.compute_states_at_sample_times(event_states=enc_out,
                                                             sample_dtimes=sample_dtimes)
        lambda_t_sample = self.softplus(state_t_sample)

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=time_delta_seqs[:, 1:],
                                                                        seq_mask=batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=type_mask[:, 1:])

        # return enc_inten to compute accuracy
        loss = - (event_ll - non_event_ll).sum()

        return loss, num_events

    def compute_states_at_sample_times(self, event_states, sample_dtimes):
        """Compute the hidden states at sampled times.

        Args:
            event_states (tensor): [batch_size, seq_len, hidden_size].
            sample_dtimes (tensor): [batch_size, seq_len, num_samples].

        Returns:
            tensor: hidden state at each sampled time.
        """
        # [batch_size, seq_len, 1, hidden_size]
        event_states = event_states[:, :, None, :]

        # [batch_size, seq_len, num_samples, 1]
        sample_dtimes = sample_dtimes[..., None]

        # [1, 1, 1, num_event_types]
        factor_intensity_decay = self.factor_intensity_decay[None, None, ...]
        factor_intensity_base = self.factor_intensity_base[None, None, ...]

        # update time decay based on Equation (6)
        # [batch_size, seq_len, num_samples, num_event_types]
        intensity_states = factor_intensity_decay * sample_dtimes + self.layer_intensity_hidden(
            event_states) + factor_intensity_base

        return intensity_states

    def compute_intensities_at_sample_times(self,
                                            time_seqs,
                                            time_delta_seqs,
                                            type_seqs,
                                            sample_dtimes,
                                            **kwargs):
        """Compute hidden states at sampled times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], times seqs.
            time_delta_seqs (tensor): [batch_size, seq_len], time delta seqs.
            type_seqs (tensor): [batch_size, seq_len], event type seqs.
            sample_dtimes (tensor): [batch_size, seq_len, num_samples], sampled inter-event timestamps.

        Returns:
            tensor: [batch_size, seq_len, num_samples, num_event_types], intensity at all sampled times.
        """

        attention_mask = kwargs.get('attention_mask', None)
        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        if attention_mask is None:
            batch_size, seq_len = time_seqs.size()
            attention_mask = torch.triu(torch.ones(seq_len, seq_len, device=self.device), diagonal=1).unsqueeze(0)
            attention_mask = attention_mask.expand(batch_size, -1, -1).to(torch.bool)

        # [batch_size, seq_len, num_samples]
        enc_out = self.forward(time_seqs, type_seqs, attention_mask)

        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = self.compute_states_at_sample_times(enc_out, sample_dtimes)

        if compute_last_step_only:
            lambdas = self.softplus(encoder_output[:, -1:, :, :])
        else:
            # [batch_size, seq_len, num_samples, num_event_types]
            lambdas = self.softplus(encoder_output)
        return lambdas



from easy_tpp.preprocess.dataset import TPPDataset
from easy_tpp.preprocess.dataset import get_data_loader
from easy_tpp.preprocess.event_tokenizer import EventTokenizer
from easy_tpp.utils import load_pickle, py_assert


class TPPDataLoader:
    def __init__(self, data_config, backend, **kwargs):
        """Initialize the dataloader

        Args:
            data_config (EasyTPP.DataConfig): data config.
            backend (str): backend engine, e.g., tensorflow or torch.
        """
        self.data_config = data_config
        self.num_event_types = data_config.data_specs.num_event_types
        self.backend = backend
        self.kwargs = kwargs

    def build_input_from_pkl(self, source_dir: str, split: str):
        data = load_pickle(source_dir)

        py_assert(data["dim_process"] == self.num_event_types,
                  ValueError,
                  "inconsistent dim_process in different splits?")

        source_data = data[split]
        time_seqs = [[x["time_since_start"] for x in seq] for seq in source_data]
        type_seqs = [[x["type_event"] for x in seq] for seq in source_data]
        time_delta_seqs = [[x["time_since_last_event"] for x in seq] for seq in source_data]

        input_dict = dict({'time_seqs': time_seqs, 'time_delta_seqs': time_delta_seqs, 'type_seqs': type_seqs})
        return input_dict

    def build_input_from_json(self, source_dir: str, split: str):
        from datasets import load_dataset
        split_ = 'validation' if split == 'dev' else split
        # load locally
        if source_dir.split('.')[-1] == 'json':
            data = load_dataset('json', data_files={split_: source_dir})
        elif source_dir.startswith('easytpp'):
            data = load_dataset(source_dir, split=split_)
        else:
            raise NotImplementedError

        py_assert(data[split_].data['dim_process'][0].as_py() == self.num_event_types,
                  ValueError,
                  "inconsistent dim_process in different splits?")

        source_data = data[split_]['event_seqs'][0]
        time_seqs, type_seqs, time_delta_seqs = [], [], []
        for k, v in source_data.items():
            cur_time_seq, cur_type_seq, cur_time_delta_seq = [], [], []
            for k_, v_ in v.items():
                cur_time_seq.append(v_['time_since_start'])
                cur_type_seq.append(v_['type_event'])
                cur_time_delta_seq.append(v_['time_since_last_event'])
            time_seqs.append(cur_time_seq)
            type_seqs.append(cur_type_seq)
            time_delta_seqs.append(cur_time_delta_seq)

        input_dict = dict({'time_seqs': time_seqs, 'time_delta_seqs': time_delta_seqs, 'type_seqs': type_seqs})
        return input_dict

    def get_loader(self, split='train', **kwargs):
        """Get the corresponding data loader.

        Args:
            split (str, optional): denote the train, valid and test set. Defaults to 'train'.
            num_event_types (int, optional): num of event types in the data. Defaults to None.

        Raises:
            NotImplementedError: the input of 'num_event_types' is inconsistent with the data.

        Returns:
            EasyTPP.DataLoader: the data loader for tpp data.
        """
        data_dir = self.data_config.get_data_dir(split)
        data_source_type = data_dir.split('.')[-1]

        if data_source_type == 'pkl':
            data = self.build_input_from_pkl(data_dir, split)
        else:
            data = self.build_input_from_json(data_dir, split)

        dataset = TPPDataset(data)
        tokenizer = EventTokenizer(self.data_config.data_specs)
        loader = get_data_loader(dataset,
                                 self.backend,
                                 tokenizer,
                                 batch_size=self.kwargs['batch_size'],
                                 shuffle=self.kwargs['shuffle'],
                                 **kwargs)

        return loader

    def train_loader(self, **kwargs):
        """Return the train loader

        Returns:
            EasyTPP.DataLoader: data loader for train set.
        """
        return self.get_loader('train', **kwargs)

    def valid_loader(self, **kwargs):
        """Return the valid loader

        Returns:
            EasyTPP.DataLoader: data loader for valid set.
        """
        return self.get_loader('dev', **kwargs)

    def test_loader(self, **kwargs):
        """Return the test loader

        Returns:
            EasyTPP.DataLoader: data loader for test set.
        """
        return self.get_loader('test', **kwargs)



from easy_tpp.hpo.base_hpo import HyperTuner
from easy_tpp.hpo.optuna_hpo import OptunaTuner
from easy_tpp.default_registers.register_optuna_trials import *

__all__ = ['HyperTuner',
           'OptunaTuner']


import tensorflow as tf
from tensorflow import keras

from easy_tpp.utils.tf_utils import get_shape_list


class EventSampler(keras.Model):
    """
    Event Sequence Sampler based on thinning algorithm

    The algorithm can be found at Algorithm 2 of The Neural Hawkes Process: A Neurally Self-Modulating
    Multivariate Point Process, https://arxiv.org/abs/1612.09328.

    The implementation uses code from https://github.com/yangalan123/anhp-andtt/blob/master/anhp/esm/thinning.py

    """

    def __init__(self, num_sample, num_exp, over_sample_rate, num_samples_boundary, dtime_max, patience_counter):
        super(EventSampler, self).__init__()
        self.num_sample = num_sample  # number of sampled next event times via thinning algorithm,
        # used to compute predictions
        self.num_exp = num_exp  # number of i.i.d. Exp(intensity_bound) draws at one time in thinning algorithm
        self.over_sample_rate = over_sample_rate
        self.num_samples_boundary = num_samples_boundary
        self.dtime_max = dtime_max
        self.patience_counter = patience_counter

    def compute_intensity_upper_bound(self,
                                      time_seqs,
                                      time_delta_seqs,
                                      type_seqs,
                                      intensity_fn,
                                      compute_last_step_only):
        """

        Args:
            time_seqs: [batch_size, seq_len]
            time_delta_seqs: [batch_size, seq_len]
            type_seqs: [batch_size, seq_len]

        Returns:
            The upper bound of intensity at each event timestamp
            [batch_size, seq_len]

        """

        # [1, 1, num_samples_boundary]
        time_for_bound_sampled = tf.linspace(start=0.0,
                                             stop=1.0,
                                             num=self.num_samples_boundary)[None, None, :]

        # [batch_size, seq_len, num_sample]
        dtime_for_bound_sampled = time_delta_seqs[:, :, None] * time_for_bound_sampled

        # [batch_size, seq_len, num_sample, event_num]
        intensities_for_bound = intensity_fn(time_seqs,
                                             time_delta_seqs,
                                             type_seqs,
                                             dtime_for_bound_sampled,
                                             compute_last_step_only=compute_last_step_only)

        # [batch_size, seq_len]
        bounds = tf.reduce_max(tf.reduce_sum(intensities_for_bound, axis=-1), axis=-1) * self.over_sample_rate

        return bounds

    def sample_exp_distribution(self, sample_rate):
        """

        Args:
            sample_rate: [batch_size, seq_len]
            time_delta_seq: [batch_size, seq_len]

        Returns:
            exp_numbers: [batch_size, seq_len, num_sample, num_exp]

        """

        # can not pass batch_size and seq_len to the random generator as they are dynamics
        # so we reuse rnd for all samples
        batch_size, seq_len = get_shape_list(sample_rate)
        # [batch_size, seq_len, num_exp]
        exp_numbers = tf.random.gamma(shape=[batch_size, seq_len, self.num_exp], alpha=1.0)

        # [batch_size, seq_len, num_exp]
        # exp_numbers = torch.tile(exp_numbers, [1, 1, self.num_sample, 1])

        # [batch_size, seq_len, num_exp]
        # div by sample_rate is equivalent to exp(sample_rate),
        # see https://en.wikipedia.org/wiki/Exponential_distribution
        exp_numbers = exp_numbers / sample_rate[:, :, None]

        return exp_numbers

    def sample_uniform_distribution(self, intensity_upper_bound):
        """

        Returns:
            unif_numbers: [batch_size, seq_len, num_sample, num_exp]

        """
        batch_size, seq_len = get_shape_list(intensity_upper_bound)
        unif_numbers = tf.random.uniform([batch_size, seq_len, self.num_sample, self.num_exp])

        return unif_numbers

    def sample_accept(self, unif_numbers, sample_rate, total_intensities):
        """

        Args:
            unif_numbers: [batch_size, max_len, num_sample, num_exp]
            sample_rate: [batch_size, max_len]
            total_intensities: [batch_size, seq_len, num_sample, num_exp]

        for each parallel draw, find its min criterion： if that < 1.0, the 1st (i.e. smallest) sampled time
        with cri < 1.0 is accepted; if none is accepted, use boundary / maxsampletime for that draw

        Returns:

        """
        # [batch_size, max_len, num_sample, num_exp]
        criterion = unif_numbers * sample_rate[:, :, None, None] / total_intensities

        # [batch_size, max_len, num_sample]
        min_cri_each_draw = tf.reduce_min(criterion, axis=-1)

        # find out unif_numbers * sample_rate < intensity
        # [batch_size, max_len, num_sample]
        who_has_accepted_times = min_cri_each_draw < 1.0

        return criterion, who_has_accepted_times

    def draw_next_time_one_step(self,
                                time_seqs,
                                time_delta_seqs,
                                type_seqs,
                                dtime_boundary,
                                intensity_fn,
                                compute_last_step_only):
        # 1. compute the upper bound of the intensity at each timestamp
        # [batch_size, seq_len]
        intensity_upper_bound = self.compute_intensity_upper_bound(time_seqs,
                                                                   time_delta_seqs,
                                                                   type_seqs,
                                                                   intensity_fn,
                                                                   compute_last_step_only)

        # 2. draw exp distribution with intensity = intensity_upper_bound
        # we apply fast approximation, i.e., re-use exp sample times for computation
        # [batch_size, seq_len, num_exp]
        exp_numbers = self.sample_exp_distribution(intensity_upper_bound)

        # 3. compute intensity at sampled times from exp distribution
        # [batch_size, seq_len, num_exp, event_num]
        intensities_at_sampled_times = intensity_fn(time_seqs,
                                                    time_delta_seqs,
                                                    type_seqs,
                                                    exp_numbers,
                                                    compute_last_step_only=compute_last_step_only)

        # [batch_size, seq_len, num_exp]
        total_intensities = tf.reduce_sum(intensities_at_sampled_times, axis=-1)

        # add one dim of num_sample: re-use the intensity for samples for prediction
        # [batch_size, seq_len, num_sample, num_exp]
        total_intensities = tf.tile(total_intensities[:, :, None, :], [1, 1, self.num_sample, 1])
        # [batch_size, seq_len, num_sample, num_exp]
        exp_numbers = tf.tile(exp_numbers[:, :, None, :], [1, 1, self.num_sample, 1])

        # 4. draw uniform distribution
        # [batch_size, seq_len, num_sample, num_exp]
        unif_numbers = self.sample_uniform_distribution(intensity_upper_bound)

        # 5. find out accepted intensities
        # [batch_size, max_len, num_sample]
        criterion, who_has_accepted_times = self.sample_accept(unif_numbers, intensity_upper_bound,
                                                               total_intensities)

        # 6. find out accepted dtimes
        sampled_dtimes_accepted = tf.identity(exp_numbers)
        # for unaccepted, use boundary/maxsampletime for that draw

        # [batch_size, seq_len, num_sample, num_exp]
        sampled_dtimes_accepted = tf.where(criterion >= 1.0,
                                           tf.ones_like(sampled_dtimes_accepted) * tf.reduce_max(exp_numbers,
                                                                                                 axis=-1,
                                                                                                 keepdims=True) + 1.0,
                                           sampled_dtimes_accepted)

        accepted_times_each_draw = tf.reduce_min(sampled_dtimes_accepted, axis=-1)

        # 7. fill out result
        # [batch_size, seq_len, num_sample]
        dtime_boundary = tf.tile(dtime_boundary[..., None], [1, 1, self.num_sample])

        # [batch_size, seq_len, num_sample]
        res = tf.ones_like(dtime_boundary) * dtime_boundary

        # [batch_size, seq_len, num_sample]
        weights = tf.ones_like(dtime_boundary)
        weights /= tf.reduce_sum(weights, axis=-1, keepdims=True)

        # [batch_size, seq_len, num_sample]
        res = tf.where(who_has_accepted_times,
                       tf.ones_like(res) * accepted_times_each_draw,
                       res)

        who_not_accept = ~who_has_accepted_times

        who_reach_further = exp_numbers[..., -1] > dtime_boundary

        res = tf.where(who_not_accept & who_reach_further,
                       exp_numbers[..., -1],
                       res)

        return res, weights



from easy_tpp.config_factory.config import Config
from easy_tpp.config_factory.data_config import DataConfig, DataSpecConfig
from easy_tpp.config_factory.hpo_config import HPOConfig, HPORunnerConfig
from easy_tpp.config_factory.runner_config import RunnerConfig, ModelConfig, BaseConfig

__all__ = ['Config',
           'DataConfig',
           'DataSpecConfig',
           'ModelConfig',
           'BaseConfig',
           'RunnerConfig',
           'HPOConfig',
           'HPORunnerConfig']



import math

import torch
from torch import nn

from easy_tpp.model.torch_model.torch_baselayer import EncoderLayer, MultiHeadAttention
from easy_tpp.model.torch_model.torch_basemodel import TorchBaseModel


class AttNHP(TorchBaseModel):
    """Torch implementation of Attentive Neural Hawkes Process, ICLR 2022.
    https://arxiv.org/abs/2201.00044.
    Source code: https://github.com/yangalan123/anhp-andtt/blob/master/anhp/model/xfmr_nhp_fast.py
    """

    def __init__(self, model_config):
        """Initialize the model

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        super(AttNHP, self).__init__(model_config)
        self.d_model = model_config.hidden_size
        self.use_norm = model_config.use_ln
        self.d_time = model_config.time_emb_size

        self.div_term = torch.exp(torch.arange(0, self.d_time, 2) * -(math.log(10000.0) / self.d_time)).reshape(1, 1,
                                                                                                                -1)

        self.n_layers = model_config.num_layers
        self.n_head = model_config.num_heads
        self.dropout = model_config.dropout_rate

        self.heads = []
        for i in range(self.n_head):
            self.heads.append(
                nn.ModuleList(
                    [EncoderLayer(
                        self.d_model + self.d_time,
                        MultiHeadAttention(1, self.d_model + self.d_time, self.d_model, self.dropout,
                                           output_linear=False),

                        use_residual=False,
                        dropout=self.dropout
                    )
                        for _ in range(self.n_layers)
                    ]
                )
            )
        self.heads = nn.ModuleList(self.heads)

        if self.use_norm:
            self.norm = nn.LayerNorm(self.d_model)
        self.inten_linear = nn.Linear(self.d_model * self.n_head, self.num_event_types)
        self.softplus = nn.Softplus()
        self.layer_event_emb = nn.Linear(self.d_model + self.d_time, self.d_model)
        self.layer_intensity = nn.Sequential(self.inten_linear, self.softplus)
        self.eps = torch.finfo(torch.float32).eps

    def compute_temporal_embedding(self, time):
        """Compute the temporal embedding.

        Args:
            time (tensor): [batch_size, seq_len].

        Returns:
            tensor: [batch_size, seq_len, emb_size].
        """
        batch_size = time.size(0)
        seq_len = time.size(1)
        pe = torch.zeros(batch_size, seq_len, self.d_time).to(time)
        _time = time.unsqueeze(-1)
        div_term = self.div_term.to(time)
        pe[..., 0::2] = torch.sin(_time * div_term)
        pe[..., 1::2] = torch.cos(_time * div_term)

        return pe

    def forward_pass(self, init_cur_layer, time_emb, sample_time_emb, event_emb, combined_mask):
        """update the structure sequentially.

        Args:
            init_cur_layer (tensor): [batch_size, seq_len, hidden_size]
            time_emb (tensor): [batch_size, seq_len, hidden_size]
            sample_time_emb (tensor): [batch_size, seq_len, hidden_size]
            event_emb (tensor): [batch_size, seq_len, hidden_size]
            combined_mask (tensor): [batch_size, seq_len, hidden_size]

        Returns:
            tensor: [batch_size, seq_len, hidden_size*2]
        """
        cur_layers = []
        seq_len = event_emb.size(1)
        for head_i in range(self.n_head):
            # [batch_size, seq_len, hidden_size]
            cur_layer_ = init_cur_layer
            for layer_i in range(self.n_layers):
                # each layer concats the temporal emb
                # [batch_size, seq_len, hidden_size*2]
                layer_ = torch.cat([cur_layer_, sample_time_emb], dim=-1)
                # make combined input from event emb + layer emb
                # [batch_size, seq_len*2, hidden_size*2]
                _combined_input = torch.cat([event_emb, layer_], dim=1)
                enc_layer = self.heads[head_i][layer_i]
                # compute the output
                enc_output = enc_layer(_combined_input, combined_mask)

                # the layer output
                # [batch_size, seq_len, hidden_size]
                _cur_layer_ = enc_output[:, seq_len:, :]
                # add residual connection
                cur_layer_ = torch.tanh(_cur_layer_) + cur_layer_

                # event emb
                event_emb = torch.cat([enc_output[:, :seq_len, :], time_emb], dim=-1)

                if self.use_norm:
                    cur_layer_ = self.norm(cur_layer_)
            cur_layers.append(cur_layer_)
        cur_layer_ = torch.cat(cur_layers, dim=-1)

        return cur_layer_

    def seq_encoding(self, time_seqs, event_seqs):
        """Encode the sequence.

        Args:
            time_seqs (tensor): time seqs input, [batch_size, seq_len].
            event_seqs (_type_): event type seqs input, [batch_size, seq_len].

        Returns:
            tuple: event embedding, time embedding and type embedding.
        """
        # [batch_size, seq_len, hidden_size]
        time_emb = self.compute_temporal_embedding(time_seqs)
        # [batch_size, seq_len, hidden_size]
        type_emb = torch.tanh(self.layer_type_emb(event_seqs.long()))
        # [batch_size, seq_len, hidden_size*2]
        event_emb = torch.cat([type_emb, time_emb], dim=-1)

        return event_emb, time_emb, type_emb

    def make_layer_mask(self, attention_mask):
        """Create a tensor to do masking on layers.

        Args:
            attention_mask (tensor): mask for attention operation, [batch_size, seq_len, seq_len]

        Returns:
            tensor: aim to keep the current layer, the same size of attention mask
            a diagonal matrix, [batch_size, seq_len, seq_len]
        """
        # [batch_size, seq_len, seq_len]
        layer_mask = (torch.eye(attention_mask.size(1)) < 1).unsqueeze(0).expand_as(attention_mask)
        return layer_mask

    def make_combined_att_mask(self, attention_mask, layer_mask):
        """Combined attention mask and layer mask.

        Args:
            attention_mask (tensor): mask for attention operation, [batch_size, seq_len, seq_len]
            layer_mask (tensor): mask for other layers, [batch_size, seq_len, seq_len]

        Returns:
            tensor: [batch_size, seq_len * 2, seq_len * 2]
        """
        # [batch_size, seq_len, seq_len * 2]
        combined_mask = torch.cat([attention_mask, layer_mask], dim=-1)
        # [batch_size, seq_len, seq_len * 2]
        contextual_mask = torch.cat([attention_mask, torch.ones_like(layer_mask)], dim=-1)
        # [batch_size, seq_len * 2, seq_len * 2]
        combined_mask = torch.cat([contextual_mask, combined_mask], dim=1)
        return combined_mask

    def forward(self, time_seqs, event_seqs, attention_mask, sample_times=None):
        """Call the model.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            event_seqs (tensor): [batch_size, seq_len], sequences of event types.
            attention_mask (tensor): [batch_size, seq_len, seq_len], masks for event sequences.
            sample_times (tensor, optional): [batch_size, seq_len, num_samples]. Defaults to None.

        Returns:
            tensor: states at sampling times, [batch_size, seq_len, num_samples].
        """
        event_emb, time_emb, type_emb = self.seq_encoding(time_seqs, event_seqs)
        init_cur_layer = torch.zeros_like(type_emb)
        layer_mask = self.make_layer_mask(attention_mask)
        if sample_times is None:
            sample_time_emb = time_emb
        else:
            sample_time_emb = self.compute_temporal_embedding(sample_times)
        combined_mask = self.make_combined_att_mask(attention_mask, layer_mask)
        cur_layer_ = self.forward_pass(init_cur_layer, time_emb, sample_time_emb, event_emb, combined_mask)

        return cur_layer_

    def loglike_loss(self, batch):
        """Compute the loglike loss.

        Args:
            batch (list): batch input.

        Returns:
            list: loglike loss, num events.
        """
        time_seqs, time_delta_seqs, type_seqs, batch_non_pad_mask, attention_mask, type_mask = batch
        # 1. compute event-loglik
        # the prediction of last event has no label, so we proceed to the last but one
        # att mask => diag is False, not mask.
        enc_out = self.forward(time_seqs[:, :-1], type_seqs[:, :-1], attention_mask[:, 1:, :-1], time_seqs[:, 1:])
        # [batch_size, seq_len, num_event_types]
        lambda_at_event = self.layer_intensity(enc_out)

        # 2. compute non-event-loglik (using MC sampling to compute integral)
        # 2.1 sample times
        # [batch_size, seq_len, num_sample]
        temp_time = self.make_dtime_loss_samples(time_delta_seqs[:, 1:])

        # [batch_size, seq_len, num_sample]
        sample_times = temp_time + time_seqs[:, :-1].unsqueeze(-1)

        # 2.2 compute intensities at sampled times
        # [batch_size, seq_len = max_len - 1, num_sample, event_num]
        lambda_t_sample = self.compute_intensities_at_sample_times(time_seqs[:, :-1],
                                                                   time_delta_seqs[:, :-1],  # not used
                                                                   type_seqs[:, :-1],
                                                                   sample_times,
                                                                   attention_mask=attention_mask[:, 1:, :-1])

        event_ll, non_event_ll, num_events = self.compute_loglikelihood(lambda_at_event=lambda_at_event,
                                                                        lambdas_loss_samples=lambda_t_sample,
                                                                        time_delta_seq=time_delta_seqs[:, 1:],
                                                                        seq_mask=batch_non_pad_mask[:, 1:],
                                                                        lambda_type_mask=type_mask[:, 1:])

        # return enc_inten to compute accuracy
        loss = - (event_ll - non_event_ll).sum()

        return loss, num_events

    def compute_states_at_sample_times(self,
                                       time_seqs,
                                       type_seqs,
                                       attention_mask,
                                       sample_times):
        """Compute the states at sampling times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
            type_seqs (tensor): [batch_size, seq_len], sequences of event types.
            attention_mask (tensor): [batch_size, seq_len, seq_len], masks for event sequences.
            sample_dtimes (tensor): delta times in sampling.

        Returns:
            tensor: hiddens states at sampling times.
        """
        batch_size = type_seqs.size(0)
        seq_len = type_seqs.size(1)
        num_samples = sample_times.size(-1)

        # [num_samples, batch_size, seq_len]
        sample_times = sample_times.permute((2, 0, 1))
        # [num_samples * batch_size, seq_len]
        _sample_time = sample_times.reshape(num_samples * batch_size, -1)
        # [num_samples * batch_size, seq_len]
        _types = type_seqs.expand(num_samples, -1, -1).reshape(num_samples * batch_size, -1)
        # [num_samples * batch_size, seq_len]
        _times = time_seqs.expand(num_samples, -1, -1).reshape(num_samples * batch_size, -1)
        # [num_samples * batch_size, seq_len]
        _attn_mask = attention_mask.unsqueeze(0).expand(num_samples, -1, -1, -1).reshape(num_samples * batch_size,
                                                                                         seq_len,
                                                                                         seq_len)
        # [num_samples * batch_size, seq_len, hidden_size]
        encoder_output = self.forward(_times,
                                      _types,
                                      _attn_mask,
                                      _sample_time)

        # [num_samples, batch_size, seq_len, hidden_size]
        encoder_output = encoder_output.reshape(num_samples, batch_size, seq_len, -1)
        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = encoder_output.permute((1, 2, 0, 3))
        return encoder_output

    def compute_intensities_at_sample_times(self, time_seqs, time_delta_seqs, type_seqs, sample_times, **kwargs):
        """Compute the intensity at sampled times.

        Args:
            time_seqs (tensor): [batch_size, seq_len], sequences of timestamps.
            time_delta_seqs (tensor): [batch_size, seq_len], sequences of delta times.
            type_seqs (tensor): [batch_size, seq_len], sequences of event types.
            sampled_dtimes (tensor): [batch_size, seq_len, num_sample], sampled time delta sequence.

        Returns:
            tensor: intensities as sampled_dtimes, [batch_size, seq_len, num_samples, event_num].
        """
        attention_mask = kwargs.get('attention_mask', None)
        compute_last_step_only = kwargs.get('compute_last_step_only', False)

        if attention_mask is None:
            batch_size, seq_len = time_seqs.size()
            attention_mask = torch.triu(torch.ones(seq_len, seq_len), diagonal=1).unsqueeze(0)
            attention_mask = attention_mask.expand(batch_size, -1, -1).to(torch.bool)

        if sample_times.size()[1] < time_seqs.size()[1]:
            # we pass sample_dtimes for last time step here
            # we do a temp solution
            # [batch_size, seq_len, num_samples]
            sample_times = time_seqs[:, :, None] + torch.tile(sample_times, [1, time_seqs.size()[1], 1])

        # [batch_size, seq_len, num_samples, hidden_size]
        encoder_output = self.compute_states_at_sample_times(time_seqs, type_seqs, attention_mask, sample_times)

        if compute_last_step_only:
            lambdas = self.layer_intensity(encoder_output[:, -1:, :, :])
        else:
            # [batch_size, seq_len, num_samples, num_event_types]
            lambdas = self.layer_intensity(encoder_output)
        return lambdas



from collections import defaultdict

import numpy as np

from easy_tpp.utils.log_utils import default_logger as logger


class MetricsHelper:
    MAXIMIZE = 'maximize'
    MINIMIZE = 'minimize'
    _registry_center = defaultdict(tuple)

    @staticmethod
    def get_metric_function(name):
        if name in MetricsHelper._registry_center:
            return MetricsHelper._registry_center[name][0]
        else:
            logger.warn(f'Metric is not found: {name}')
            return None

    @staticmethod
    def get_metric_direction(name):
        if name in MetricsHelper._registry_center:
            return MetricsHelper._registry_center[name][1]
        else:
            return None

    @staticmethod
    def get_all_registered_metric():
        return MetricsHelper._registry_center.values

    @staticmethod
    def register(name, direction, overwrite=True):
        registry_center = MetricsHelper._registry_center

        def _add_metric_to_registry(func):
            if name in registry_center:
                if overwrite:
                    registry_center[name] = (func, direction)
                else:
                    logger.warn(f'The metric {name} is already registered, and cannot be overwritten!')
            else:
                registry_center[name] = (func, direction)
            return func

        return _add_metric_to_registry

    @staticmethod
    def metrics_dict_to_str(metrics_dict):
        """ Convert metrics to a string to show in console  """
        eval_info = ''
        for k, v in metrics_dict.items():
            eval_info += '{0} is {1}, '.format(k, v)

        return eval_info[:-2]

    @staticmethod
    def get_metrics_callback_from_names(metric_names):
        """ Metrics function callbacks    """
        metric_functions = []
        metric_names_ = []
        for name in metric_names:
            metric = MetricsHelper.get_metric_function(name)
            if metric is not None:
                metric_functions.append(metric)
                metric_names_.append(name)

        def metrics(preds, labels, **kwargs):
            """ call metrics functions """
            res = dict()
            for metric_name, metric_func in zip(metric_names_, metric_functions):
                res[metric_name.lower()] = metric_func(preds, labels, **kwargs)
            return res

        return metrics


class MetricsTracker:
    """Track and record the metrics.
    """

    def __init__(self):
        self.current_best = {
            'loglike': np.finfo(float).min,
            'distance': np.finfo(float).max
        }
        self.episode_best = 'NeverUpdated'

    def update_best(self, key, value, epoch):
        """Update the recorder for the best metrics.

        Args:
            key (str): metrics key.
            value (float): metrics value.
            epoch (int): num of epoch.

        Raises:
            NotImplementedError: for keys other than 'loglike'.

        Returns:
            bool: whether the recorder has been updated.
        """
        updated = False
        if key == 'loglike':
            if value > self.current_best[key]:
                updated = True
                self.current_best[key] = value
                self.episode_best = epoch
        else:
            raise NotImplementedError

        return updated



""" Base model with common functionality  """

import torch
from torch import nn

from easy_tpp.model.torch_model.torch_thinning import EventSampler
from easy_tpp.utils import set_device


class TorchBaseModel(nn.Module):
    def __init__(self, model_config):
        """Initialize the BaseModel

        Args:
            model_config (EasyTPP.ModelConfig): model spec of configs
        """
        super(TorchBaseModel, self).__init__()
        self.loss_integral_num_sample_per_step = model_config.loss_integral_num_sample_per_step
        self.hidden_size = model_config.hidden_size
        self.num_event_types = model_config.num_event_types  # not include [PAD], [BOS], [EOS]
        self.num_event_types_pad = model_config.num_event_types_pad  # include [PAD], [BOS], [EOS]
        self.pad_token_id = model_config.pad_token_id
        self.eps = torch.finfo(torch.float32).eps

        self.layer_type_emb = nn.Embedding(self.num_event_types_pad,  # have padding
                                           self.hidden_size,
                                           padding_idx=self.pad_token_id)

        self.gen_config = model_config.thinning
        self.event_sampler = None
        self.device = set_device(model_config.gpu)

        self.to(self.device)

        if self.gen_config:
            self.event_sampler = EventSampler(num_sample=self.gen_config.num_sample,
                                              num_exp=self.gen_config.num_exp,
                                              over_sample_rate=self.gen_config.over_sample_rate,
                                              patience_counter=self.gen_config.patience_counter,
                                              num_samples_boundary=self.gen_config.num_samples_boundary,
                                              dtime_max=self.gen_config.dtime_max,
                                              device=self.device)

    @staticmethod
    def generate_model_from_config(model_config):
        """Generate the model in derived class based on model config.

        Args:
            model_config (EasyTPP.ModelConfig): config of model specs.
        """
        model_id = model_config.model_id

        for subclass in TorchBaseModel.__subclasses__():
            if subclass.__name__ == model_id:
                return subclass(model_config)

        raise RuntimeError('No model named ' + model_id)

    @staticmethod
    def get_logits_at_last_step(logits, batch_non_pad_mask, sample_len=None):
        """Retrieve the hidden states of last non-pad events.

        Args:
            logits (tensor): [batch_size, seq_len, hidden_dim], a sequence of logits
            batch_non_pad_mask (tensor): [batch_size, seq_len], a sequence of masks
            sample_len (tensor): default None, use batch_non_pad_mask to find out the last non-mask position

        ref: https://medium.com/analytics-vidhya/understanding-indexing-with-pytorch-gather-33717a84ebc4

        Returns:
            tensor: retrieve the logits of EOS event
        """

        seq_len = batch_non_pad_mask.sum(dim=1)
        select_index = seq_len - 1 if sample_len is None else seq_len - 1 - sample_len
        # [batch_size, hidden_dim]
        select_index = select_index.unsqueeze(1).repeat(1, logits.size(-1))
        # [batch_size, 1, hidden_dim]
        select_index = select_index.unsqueeze(1)
        # [batch_size, hidden_dim]
        last_logits = torch.gather(logits, dim=1, index=select_index).squeeze(1)
        return last_logits

    def compute_loglikelihood(self, time_delta_seq, lambda_at_event, lambdas_loss_samples, seq_mask,
                              lambda_type_mask):
        """Compute the loglikelihood of the event sequence based on Equation (8) of NHP paper.

        Args:
            time_delta_seq (tensor): [batch_size, seq_len], time_delta_seq from model input.
            lambda_at_event (tensor): [batch_size, seq_len, num_event_types], unmasked intensity at
            (right after) the event.
            lambdas_loss_samples (tensor): [batch_size, seq_len, num_sample, num_event_types],
            intensity at sampling times.
            seq_mask (tensor): [batch_size, seq_len], sequence mask vector to mask the padded events.
            lambda_type_mask (tensor): [batch_size, seq_len, num_event_types], type mask matrix to mask the
            padded event types.

        Returns:
            tuple: event loglike, non-event loglike, intensity at event with padding events masked
        """

        # Sum of lambda over every type and every event point
        # [batch_size, seq_len]
        event_lambdas = torch.sum(lambda_at_event * lambda_type_mask, dim=-1) + self.eps

        # mask the pad event
        event_lambdas = event_lambdas.masked_fill_(~seq_mask, 1.0)

        # [batch_size, seq_len)
        event_ll = torch.log(event_lambdas)

        # Compute the big lambda integral in equation (8) of NHP paper
        # 1 - take num_mc_sample rand points in each event interval
        # 2 - compute its lambda value for every sample point
        # 3 - take average of these sample points
        # 4 - times the interval length

        # [batch_size, seq_len, n_loss_sample]
        lambdas_total_samples = lambdas_loss_samples.sum(dim=-1)

        # interval_integral - [batch_size, seq_len]
        # interval_integral = length_interval * average of sampled lambda(t)
        non_event_ll = lambdas_total_samples.mean(dim=-1) * time_delta_seq * seq_mask

        num_events = torch.masked_select(event_ll, event_ll.ne(0.0)).size()[0]

        return event_ll, non_event_ll, num_events

    def make_dtime_loss_samples(self, time_delta_seq):
        """Generate the time point samples for every interval.

        Args:
            time_delta_seq (tensor): [batch_size, seq_len].

        Returns:
            tensor: [batch_size, seq_len, n_samples]
        """
        # [1, 1, n_samples]
        dtimes_ratio_sampled = torch.linspace(start=0.0,
                                              end=1.0,
                                              steps=self.loss_integral_num_sample_per_step,
                                              device=self.device)[None, None, :]

        # [batch_size, max_len, n_samples]
        sampled_dtimes = time_delta_seq[:, :, None] * dtimes_ratio_sampled

        return sampled_dtimes

    def compute_states_at_sample_times(self, **kwargs):
        raise NotImplementedError('This need to implemented in inherited class ! ')

    def predict_one_step_at_every_event(self, batch):
        """One-step prediction for every event in the sequence.

        Args:
            time_seqs (tensor): [batch_size, seq_len].
            time_delta_seqs (tensor): [batch_size, seq_len].
            type_seqs (tensor): [batch_size, seq_len].

        Returns:
            tuple: tensors of dtime and type prediction, [batch_size, seq_len].
        """
        time_seq, time_delta_seq, event_seq, batch_non_pad_mask, _, type_mask = batch

        # remove the last event, as the prediction based on the last event has no label
        # time_delta_seq should start from 1, because the first one is zero
        time_seq, time_delta_seq, event_seq = time_seq[:, :-1], time_delta_seq[:, 1:], event_seq[:, :-1]

        # [batch_size, seq_len]
        dtime_boundary = time_delta_seq + self.event_sampler.dtime_max

        # [batch_size, seq_len, num_sample]
        accepted_dtimes, weights = self.event_sampler.draw_next_time_one_step(time_seq,
                                                                              time_delta_seq,
                                                                              event_seq,
                                                                              dtime_boundary,
                                                                              self.compute_intensities_at_sample_times)

        # [batch_size, seq_len]
        dtimes_pred = torch.sum(accepted_dtimes * weights, dim=-1)

        # [batch_size, seq_len, 1, event_num]
        intensities_at_times = self.compute_intensities_at_sample_times(time_seq,
                                                                        time_delta_seq,
                                                                        event_seq,
                                                                        dtimes_pred[:, :, None],
                                                                        max_steps=event_seq.size()[1])

        # [batch_size, seq_len, event_num]
        intensities_at_times = intensities_at_times.squeeze(dim=-2)

        types_pred = torch.argmax(intensities_at_times, dim=-1)

        return dtimes_pred, types_pred

    def predict_multi_step_since_last_event(self, batch, forward=False):
        """Multi-step prediction since last event in the sequence.

        Args:
            time_seqs (tensor): [batch_size, seq_len].
            time_delta_seqs (tensor): [batch_size, seq_len].
            type_seqs (tensor): [batch_size, seq_len].
            num_step (int): num of steps for prediction.

        Returns:
            tuple: tensors of dtime and type prediction, [batch_size, seq_len].
        """
        time_seq_label, time_delta_seq_label, event_seq_label, batch_non_pad_mask_label, _, type_mask_label = batch

        num_step = self.gen_config.num_step_gen

        if not forward:
            time_seq = time_seq_label[:, :-num_step]
            time_delta_seq = time_delta_seq_label[:, :-num_step]
            event_seq = event_seq_label[:, :-num_step]
        else:
            time_seq, time_delta_seq, event_seq = time_seq_label, time_delta_seq_label, event_seq_label

        for i in range(num_step):
            # [batch_size, seq_len]
            dtime_boundary = time_delta_seq + self.event_sampler.dtime_max

            # [batch_size, 1, num_sample]
            accepted_dtimes, weights = \
                self.event_sampler.draw_next_time_one_step(time_seq,
                                                           time_delta_seq,
                                                           event_seq,
                                                           dtime_boundary,
                                                           self.compute_intensities_at_sample_times,
                                                           compute_last_step_only=True)

            # [batch_size, 1]
            dtimes_pred = torch.sum(accepted_dtimes * weights, dim=-1)

            # [batch_size, seq_len, 1, event_num]
            intensities_at_times = self.compute_intensities_at_sample_times(time_seq,
                                                                            time_delta_seq,
                                                                            event_seq,
                                                                            dtimes_pred[:, :, None],
                                                                            max_steps=event_seq.size()[1])

            # [batch_size, seq_len, event_num]
            intensities_at_times = intensities_at_times.squeeze(dim=-2)

            # [batch_size, seq_len]
            types_pred = torch.argmax(intensities_at_times, dim=-1)

            # [batch_size, 1]
            types_pred_ = types_pred[:, -1:]
            dtimes_pred_ = dtimes_pred[:, -1:]
            time_pred_ = time_seq[:, -1:] + dtimes_pred_

            # concat to the prefix sequence
            time_seq = torch.cat([time_seq, time_pred_], dim=-1)
            time_delta_seq = torch.cat([time_delta_seq, dtimes_pred_], dim=-1)
            event_seq = torch.cat([event_seq, types_pred_], dim=-1)

        return time_delta_seq[:, -num_step - 1:], event_seq[:, -num_step - 1:], \
               time_delta_seq_label[:, -num_step - 1:], event_seq_label[:, -num_step - 1:]
